"""
A message publisher and listener for native Python objects and MXNet tensors.

This script demonstrates the capability to transmit native Python objects and MXNet tensors using
the MiddlewareCommunicator within the Wrapyfi library. The communication follows the PUB/SUB pattern
allowing message publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using the NativeObject message
    - Transmitting a nested dummy Python object with native objects and MXNet tensors
    - Applying the PUB/SUB pattern with mirroring
    - Transmitting MXNet tensors with different contexts (CPU and GPU if available)
    - Flipping of contexts by mapping CPU to GPU and vice versa

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - MXNet: Used for handling and creating tensors (refer to https://mxnet.apache.org/get_started for installation instructions)

    Install using pip:
        ``pip install mxnet``  # Basic installation of MXNet

Run:
    # On machine 1 (or process 1): Publisher waits for keyboard input and transmits message
        ``python3 mxnet_example.py --mode publish``

    # On machine 2 (or process 2): Listener waits for message and prints the entire dummy object
        ``python3 mxnet_example.py --mode listen``
"""

import argparse

try:
    import mxnet
except ImportError:
    print("Install MXNet before running this script.")

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class Notifier(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "NativeObject", "$mware", "Notifier", "/notify/test_mxnet_exchange",
        carrier="", should_wait=True,
        listener_kwargs=dict(load_mxnet_device=mxnet.gpu(0), map_mxnet_devices={'cpu': 'cuda:0', 'gpu:0': 'cpu'})
    )
    def exchange_object(self, mware=None):
        """Exchange messages with MXNet tensors and other native Python objects."""
        msg = input("Type your message: ")
        ret = {
            "message": msg,
            "mx_ones": mxnet.nd.ones((2, 4), ctx=mxnet.cpu()),
            "mxnet_zeros_cuda": mxnet.nd.zeros((2, 3), ctx=mxnet.gpu(0))
        }
        return ret,


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="A message publisher and listener for native Python objects and MXNet tensors.")
    parser.add_argument(
        "--mode", type=str, default="publish",
        choices={"publish", "listen"},
        help="The transmission mode"
    )
    parser.add_argument(
        "--mware", type=str, default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission"
    )
    return parser.parse_args()


def main(args):
    """Main function to initiate Notifier class and communication."""
    notifier = Notifier()
    notifier.activate_communication(Notifier.exchange_object, mode=args.mode)

    while True:
        msg_object, = notifier.exchange_object(mware=args.mware)
        print("Method result:", msg_object)


if __name__ == "__main__":
    args = parse_args()
    main(args)



import logging

from wrapyfi.connect.publishers import Publisher, Publishers


@Publishers.register("MMO", "fallback")
class FallbackPublisher(Publisher):

    def __init__(self, name: str, out_topic: str, carrier: str = "",
                 should_wait: bool = True, missing_middleware_object: str = "", **kwargs):
        logging.warning(f"Fallback publisher employed due to missing middleware or object type: "
                        f"{missing_middleware_object}")
        Publisher.__init__(self, name, out_topic, carrier=carrier, should_wait=should_wait, **kwargs)
        self.missing_middleware_object = missing_middleware_object

    def establish(self, repeats: int = -1, **kwargs):
        return None

    def publish(self, obj):
        return obj

    def close(self):
        return None


import logging
import sys
import json
import time
import os
import importlib.util
import queue
from typing import Optional, Tuple

import numpy as np
import rospy
import std_msgs.msg
import sensor_msgs.msg

from wrapyfi.connect.servers import Server, Servers
from wrapyfi.middlewares.ros import ROSMiddleware, ROSNativeObjectService, ROSImageService
from wrapyfi.encoders import JsonEncoder, JsonDecodeHook


class ROSServer(Server):

    def __init__(self, name: str, out_topic: str, carrier: str = "tcp", ros_kwargs: Optional[dict] = None, **kwargs):
        """
        Initialize the server.

        :param name: str: Name of the server
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for REQ/REP pattern. Default is 'tcp'
        :param ros_kwargs: dict: Additional kwargs for the ROS middleware
        :param kwargs: dict: Additional kwargs for the server
        """
        if carrier or carrier != "tcp":
            logging.warning("[ROS] ROS does not support other carriers than TCP for REQ/REP pattern. Using TCP.")
            carrier = "tcp"
        super().__init__(name, out_topic, carrier=carrier, **kwargs)
        ROSMiddleware.activate(**ros_kwargs or {})

    def close(self):
        """
        Close the server.
        """
        if hasattr(self, "_server") and self._server:
            if self._server is not None:
                self._server.shutdown()

    def __del__(self):
        self.close()


@Servers.register("NativeObject", "ros")
class ROSNativeObjectServer(ROSServer):
    SEND_QUEUE = queue.Queue(maxsize=1)
    RECEIVE_QUEUE = queue.Queue(maxsize=1)

    def __init__(self, name: str, out_topic: str, carrier: str = "tcp",
                 serializer_kwargs: Optional[dict] = None, deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        Specific server handling native Python objects, serializing them to JSON strings for transmission.

        :param name: str: Name of the server
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for REQ/REP pattern. Default is 'tcp'
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, out_topic, carrier=carrier, **kwargs)

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        self._server = None

    def establish(self):
        """
        Establish the connection to the server.
        """
        self._server = rospy.Service(self.out_topic, ROSNativeObjectService, self._service_callback)
        self.established = True

    def await_request(self, *args, **kwargs):
        """
        Await and deserialize the client's request, returning the extracted arguments and keyword arguments.
        The method blocks until a message is received, then attempts to deserialize it using the configured JSON decoder
        hook, returning the extracted arguments and keyword arguments.

        :return: Tuple[list, dict]: A tuple containing two items:
                    - A list of arguments extracted from the received message
                    - A dictionary of keyword arguments extracted from the received message
        """
        if not self.established:
            self.establish()
        try:
            request = ROSNativeObjectServer.RECEIVE_QUEUE.get(block=True)
            [args, kwargs] = json.loads(request.data, object_hook=self._plugin_decoder_hook, **self._deserializer_kwargs)
            return args, kwargs
        except rospy.ServiceException as e:
            logging.error("[ROS] Service call failed: %s" % e)
            return [], {}

    @staticmethod
    def _service_callback(msg):
        """
        Callback for the ROS service.

        :param msg: ROSNativeObjectService._request_class: The request message
        :return: ROSNativeObjectService._response_class: The response message
        """
        ROSNativeObjectServer.RECEIVE_QUEUE.put(msg)
        return ROSNativeObjectServer.SEND_QUEUE.get(block=True)

    def reply(self, obj):
        """
        Serialize the provided object and send it as a reply to the client.

        :param obj: Any: The Python object to be serialized and sent
        """
        try:
            obj_str = json.dumps(obj, cls=self._plugin_encoder, **self._plugin_kwargs,
                                 serializer_kwrags=self._serializer_kwargs)
            obj_msg = std_msgs.msg.String()
            obj_msg.data = obj_str
            ROSNativeObjectServer.SEND_QUEUE.put(obj_msg, block=False)
        except queue.Full:
            logging.warning(f"[ROS] Discarding data because queue is full. "
                            f"This happened due to bad synchronization in {self.__name__}")


@Servers.register("Image", "ros")
class ROSImageServer(ROSServer):
    SEND_QUEUE = queue.Queue(maxsize=1)
    RECEIVE_QUEUE = queue.Queue(maxsize=1)

    def __init__(self, name: str, out_topic: str, carrier: str = "tcp",
                 width: int = -1, height: int = -1, rgb: bool = True, fp: bool = False,
                 deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        Specific server handling native Python objects, serializing them to JSON strings for transmission.

        :param name: str: Name of the server
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for REQ/REP pattern. Default is 'tcp'
        :param width: int: Width of the image. Default is -1 (use the width of the received image)
        :param height: int: Height of the image. Default is -1 (use the height of the received image)
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, out_topic, carrier=carrier, **kwargs)
        if "jpg" in kwargs:
            logging.warning("[ROS] ROS currently does not support JPG encoding in REQ/REP. Using raw image.")
            kwargs.pop("jpg")
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp

        if self.fp:
            self._encoding = '32FC3' if self.rgb else '32FC1'
            self._type = np.float32
        else:
            self._encoding = 'bgr8' if self.rgb else 'mono8'
            self._type = np.uint8

        self._plugin_kwargs = kwargs
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        self._server = None

    def establish(self):
        """
        Establish the connection to the server.
        """
        self._server = rospy.Service(self.out_topic, ROSImageService, self._service_callback)
        self.established = True

    def await_request(self, *args, **kwargs):
        """
        Await and deserialize the client's request, returning the extracted arguments and keyword arguments.
        The method blocks until a message is received, then attempts to deserialize it using the configured JSON decoder
        hook, returning the extracted arguments and keyword arguments.

        :return: Tuple[list, dict]: A tuple containing two items:
                    - A list of arguments extracted from the received message
                    - A dictionary of keyword arguments extracted from the received message
        """
        if not self.established:
            self.establish()
        try:
            request = ROSImageServer.RECEIVE_QUEUE.get(block=True)
            [args, kwargs] = json.loads(request.data, object_hook=self._plugin_decoder_hook, **self._deserializer_kwargs)
            return args, kwargs
        except rospy.ServiceException as e:
            logging.error("[ROS] Service call failed: %s" % e)
            return [], {}

    @staticmethod
    def _service_callback(msg):
        """
        Callback for the ROS service.

        :param msg: ROSImageService._request_class: The request message
        :return: ROSImageService._response_class: The response message
        """
        ROSImageServer.RECEIVE_QUEUE.put(msg)
        return ROSImageServer.SEND_QUEUE.get(block=True)

    def reply(self, img: np.ndarray):
        """
        Serialize the provided image and send it as a reply to the client.

        :param img: np.ndarray: Image to publish
        """
        try:
            if 0 < self.width != img.shape[1] or 0 < self.height != img.shape[0] or \
                    not ((img.ndim == 2 and not self.rgb) or (img.ndim == 3 and self.rgb and img.shape[2] == 3)):
                raise ValueError("Incorrect image shape for publisher")
            img = np.require(img, dtype=self._type, requirements='C')
            img_msg = sensor_msgs.msg.Image()
            img_msg.header.stamp = rospy.Time.now()
            img_msg.height = img.shape[0]
            img_msg.width = img.shape[1]
            img_msg.encoding = self._encoding
            img_msg.is_bigendian = img.dtype.byteorder == '>' or (img.dtype.byteorder == '=' and sys.byteorder == 'big')
            img_msg.step = img.strides[0]
            img_msg.data = img.tobytes()
            ROSImageServer.SEND_QUEUE.put(img_msg, block=False)
        except queue.Full:
            logging.warning(f"[ROS] Discarding data because queue is full. "
                            f"This happened due to bad synchronization in {self.__name__}")


@Servers.register("AudioChunk", "ros")
class ROSAudioChunkServer(ROSServer):
    SEND_QUEUE = queue.Queue(maxsize=1)
    RECEIVE_QUEUE = queue.Queue(maxsize=1)

    def __init__(self, name: str, out_topic: str, carrier: str = "tcp",
                 channels: int = 1, rate: int = 44100, chunk: int = -1,
                 deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        Specific server handling audio data as numpy arrays.

        :param name: str: Name of the server
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for REQ/REP pattern. Default is 'tcp'
        :param channels: int: Number of channels in the audio. Default is 1
        :param rate: int: Sampling rate of the audio. Default is 44100
        :param chunk: int: Number of samples in the audio chunk. Default is -1 (use the chunk size of the received audio)
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, out_topic, carrier=carrier, **kwargs)
        self._plugin_kwargs = kwargs
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        self._server = self._rep_msg = None

        self.channels = channels
        self.rate = rate
        self.chunk = chunk

    def establish(self):
        """
        Establish the connection to the server.
        """
        try:
            from wrapyfi_ros_interfaces.srv import ROSAudioService
        except ImportError:
            import wrapyfi
            logging.error("[ROS] Could not import ROSAudioService. "
                          "Make sure the ROS services in wrapyfi_extensions/wrapyfi_ros_interfaces are compiled. "
                          "Refer to the documentation for more information: \n" +
                          wrapyfi.__doc__ + "ros_interfaces_lnk.html")
            sys.exit(1)
        self._server = rospy.Service(self.out_topic, ROSAudioService, self._service_callback)
        self._rep_msg = ROSAudioService._response_class().response
        self.established = True

    def await_request(self, *args, **kwargs):
        """
        Await and deserialize the client's request, returning the extracted arguments and keyword arguments.
        The method blocks until a message is received, then attempts to deserialize it using the configured JSON decoder
        hook, returning the extracted arguments and keyword arguments.

        :return: Tuple[list, dict]: A tuple containing two items:
                 - A list of arguments extracted from the received message
                 - A dictionary of keyword arguments extracted from the received message
        """
        if not self.established:
            self.establish()
        try:
            request = ROSAudioChunkServer.RECEIVE_QUEUE.get(block=True)
            [args, kwargs] = json.loads(request.request, object_hook=self._plugin_decoder_hook,
                                        **self._deserializer_kwargs)
            return args, kwargs
        except rospy.ServiceException as e:
            logging.error("[ROS] Service call failed: %s" % e)
            return [], {}

    @staticmethod
    def _service_callback(msg):
        """
        Callback for the ROS service.

        :param msg: ROSAudioService._request_class: The request message
        :return: ROSAudioService._response_class: The response message
        """
        ROSAudioChunkServer.RECEIVE_QUEUE.put(msg)
        return ROSAudioChunkServer.SEND_QUEUE.get(block=True)

    def reply(self, aud: Tuple[np.ndarray, int]):
        """
        Serialize the provided audio data and send it as a reply to the client.

        :param aud: Tuple[np.ndarray, int]: Audio chunk to publish formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        try:
            aud, rate = aud
            if aud is None:
                return
            if 0 < self.rate != rate:
                raise ValueError("Incorrect audio rate for publisher")
            chunk, channels = aud.shape if len(aud.shape) > 1 else (aud.shape[0], 1)
            self.chunk = chunk if self.chunk == -1 else self.chunk
            self.channels = channels if self.channels == -1 else self.channels
            if 0 < self.chunk != chunk or 0 < self.channels != channels:
                raise ValueError("Incorrect audio shape for publisher")
            aud = np.require(aud, dtype=np.float32, requirements='C')

            aud_msg = self._rep_msg
            aud_msg.header.stamp = rospy.Time.now()
            aud_msg.chunk_size = chunk
            aud_msg.channels = channels
            aud_msg.sample_rate = rate
            aud_msg.is_bigendian = aud.dtype.byteorder == '>' or (aud.dtype.byteorder == '=' and sys.byteorder == 'big')
            aud_msg.encoding = 'S16BE' if aud_msg.is_bigendian else 'S16LE'
            aud_msg.step = aud.strides[0]
            aud_msg.data = aud.tobytes()  # (aud * 32767.0).tobytes()
            ROSAudioChunkServer.SEND_QUEUE.put(aud_msg, block=False)
        except queue.Full:
            logging.warning(f"[ROS] Discarding data because queue is full. "
                            f"This happened due to bad synchronization in {self.__name__}")



"""
Encoder and Decoder for JAX Tensor Data via Wrapyfi.

This script provides mechanisms to encode and decode JAX tensor data using Wrapyfi.
It utilizes base64 encoding to convert binary data into ASCII strings.

The script contains a class, `JAXTensor`, registered as a plugin to manage the
conversion of JAX tensor data (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - JAX: An open-source high-performance machine learning library (refer to https://github.com/google/jax for installation instructions)
        Note: If JAX is not available, HAVE_JAX will be set to False and
        the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install jax jaxlib``  # Basic installation of JAX
"""

import io
import base64

import numpy as np

from wrapyfi.utils import *

try:
    import jax
    HAVE_JAX = True
except ImportError:
    HAVE_JAX = False

try:
    # if jax 0.3.22 is installed, then jax.numpy is a module
    types = None if not HAVE_JAX else jax.numpy.DeviceArray.__mro__[:-1] + (jax.Array,)
except AttributeError:
    types = None if not HAVE_JAX else jax.numpy.DeviceArray.__mro__[:-1]


@PluginRegistrar.register(types=types)
class JAXTensor(Plugin):
    def __init__(self, **kwargs):
        """
        Initialize the JAXTensor plugin.
        """
        pass

    def encode(self, obj, *args, **kwargs):
        """
        Encode JAX tensor data into a base64 ASCII string.

        :param obj: jax.numpy.DeviceArray: The JAX tensor data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name and encoded data string
        """
        with io.BytesIO() as memfile:
            np.save(memfile, np.asarray(obj))
            obj_data = base64.b64encode(memfile.getvalue()).decode('ascii')
        return True, dict(__wrapyfi__=(str(self.__class__.__name__), obj_data))

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a base64 ASCII string back into JAX tensor data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the encoded data string
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, jax.numpy.DeviceArray]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - jax.numpy.DeviceArray: The decoded JAX tensor data
        """
        with io.BytesIO(base64.b64decode(obj_full[1].encode('ascii'))) as memfile:
            return True, jax.numpy.array(np.load(memfile))




import logging
import atexit

import yarp

from wrapyfi.utils import SingletonOptimized
from wrapyfi.connect.wrapper import MiddlewareCommunicator


class YarpMiddleware(metaclass=SingletonOptimized):
    """
    YARP middleware wrapper. This class is a singleton, so it can be instantiated only once. The ``activate`` method
    should be called to initialize the middleware. The ``deinit`` method should be called to deinitialize the middleware
    and destroy all connections. The ``activate`` and ``deinit`` methods are automatically called when the class is
    instantiated and when the program exits, respectively.
    """
    @staticmethod
    def activate(**kwargs):
        """
        Activate the YARP middleware. This method should be called to initialize the middleware.

        :param kwargs: dict: Keyword arguments to be passed to the YARP initialization function
        """
        YarpMiddleware(**kwargs)

    def __init__(self, *args, **kwargs):
        """
        Initialize the YARP middleware. This method is automatically called when the class is instantiated.
        """
        logging.info("Initialising YARP middleware")
        yarp.Network.init()
        atexit.register(MiddlewareCommunicator.close_all_instances)
        atexit.register(self.deinit)

    @staticmethod
    def deinit():
        """
        Deinitialize the YARP middleware. This method is automatically called when the program exits.
        """
        logging.info("Deinitializing YARP middleware")
        yarp.Network.fini()



import unittest


class ZeroMQTestWrapper(unittest.TestCase):
    MWARE = "zeromq"

    def test_activate_communication(self):
        """
        Test the activate communication functionality of the middleware. When the ``activate_communication`` method is
        called, the ``__WRAPYFI_INSTANCES`` attribute of the decorated function should be updated with the instance of
        the class that called the method. Destroying the class instance should remove the instance from the list and the
        the connection should no longer be active. The ``activate_communication`` method should also be callable multiple times on the same function.
        """
        import wrapyfi.tests.tools.class_test as class_test
        Test = class_test.Test
        Test.close_all_instances()
        if self.MWARE not in Test.get_communicators():
            self.skipTest(f"{self.MWARE} not installed")

        # create three node nodes to test closing and auto-naming of the second node
        test, test2, test3 = Test(), Test(), Test()
        test.activate_communication(test.exchange_object, mode="publish")
        test2.activate_communication(test2.exchange_object, mode=None)
        test2.activate_communication(test2.exchange_object, mode="publish")  # override None
        test3.activate_communication(test3.exchange_object, mode="disable")

        # ensure both nodes are registered
        self.assertIn("Test.exchange_object", Test._MiddlewareCommunicator__registry)
        self.assertIn("Test.exchange_object.2", Test._MiddlewareCommunicator__registry)
        self.assertIn("Test.exchange_object.3", Test._MiddlewareCommunicator__registry)
        for i in range(2):
            msg_object, = test.exchange_object(mware=self.MWARE)
            msg_object2, = test2.exchange_object(mware=self.MWARE, topic="/test/test_native_exchange2")
            msg_object3, = test3.exchange_object(mware=self.MWARE)
            self.assertDictEqual(msg_object, msg_object2)
            self.assertIsNone(msg_object3)
        test.close()
        del test

        # ensure only one node is registered
        self.assertIn("Test.exchange_object", Test._MiddlewareCommunicator__registry)
        self.assertIn("Test.exchange_object.2", Test._MiddlewareCommunicator__registry)
        self.assertNotIn("Test.exchange_object.3", Test._MiddlewareCommunicator__registry)

        for i in range(2):
            msg_object2, = test2.exchange_object(mware=self.MWARE, topic="/test/test_native_exchange2")
        test2.close()
        test3.close()
        del test2, test3

        # ensure no nodes are registered
        self.assertIn("Test.exchange_object", Test._MiddlewareCommunicator__registry)

    def test_close(self):
        """
        Test the close functionality of the middleware. When the ``close`` method is called, the instances should be reordered
        in the ``__WRAPYFI_INSTANCES`` attribute of the decorated function. Destroying the class instance should remove the
        instance from the list and the index of the connection should change. The ``close`` method should also be callable
        multiple times on different instances of a class.
        """
        import wrapyfi.tests.tools.class_test as class_test
        Test = class_test.Test
        Test.close_all_instances()

        if self.MWARE not in Test.get_communicators():
            self.skipTest(f"{self.MWARE} not installed")

        # create three nodes to test reordering on closing
        test, test2, test3 = Test(), Test(), Test()
        test.activate_communication(test.exchange_object, mode="disable")
        test2.activate_communication(test2.exchange_object, mode=None)
        test3.activate_communication(test3.exchange_object, mode="publish")

        # ensure all nodes are registered
        self.assertIn("Test.exchange_object", Test._MiddlewareCommunicator__registry)
        self.assertIn("Test.exchange_object.2", Test._MiddlewareCommunicator__registry)
        self.assertIn("Test.exchange_object.3", Test._MiddlewareCommunicator__registry)

        # delete the second node. The third node should be renamed to second
        test2.close()
        del test2
        self.assertIn("Test.exchange_object", Test._MiddlewareCommunicator__registry)
        self.assertIn("Test.exchange_object.2", Test._MiddlewareCommunicator__registry)
        self.assertNotIn("Test.exchange_object.3", Test._MiddlewareCommunicator__registry)

        # check that the first & third node are still registered
        test_id = Test._MiddlewareCommunicator__registry["Test.exchange_object"]["__WRAPYFI_INSTANCES"].index(hex(id(test)))
        test3_id = Test._MiddlewareCommunicator__registry["Test.exchange_object"]["__WRAPYFI_INSTANCES"].index(hex(id(test3)))
        self.assertEqual(test_id, 0)
        self.assertEqual(test3_id, 1)

        # close the two nodes
        test.close()
        del test

        # recreate test node which should be renamed to second, and the third node should be renamed to first (no index)
        test = Test()
        test.activate_communication(test.exchange_object, mode="publish")
        self.assertIn("Test.exchange_object", Test._MiddlewareCommunicator__registry)

        # check that the first & third node are still registered
        test_id = Test._MiddlewareCommunicator__registry["Test.exchange_object"]["__WRAPYFI_INSTANCES"].index(hex(id(test)))
        test3_id = Test._MiddlewareCommunicator__registry["Test.exchange_object"]["__WRAPYFI_INSTANCES"].index(hex(id(test3)))
        self.assertEqual(test_id, 1)
        self.assertEqual(test3_id, 0)

        # ensure no nodes are registered
        test3.close()
        del test3
        test.close()
        del test

        self.assertIn("Test.exchange_object", Test._MiddlewareCommunicator__registry)
        self.assertNotIn("Test.exchange_object.2", Test._MiddlewareCommunicator__registry)
        self.assertNotIn("Test.exchange_object.3", Test._MiddlewareCommunicator__registry)

        # close all instances after all have been closed
        Test.close_all_instances()
        self.assertIn("Test.exchange_object", Test._MiddlewareCommunicator__registry)
        self.assertNotIn("Test.exchange_object.2", Test._MiddlewareCommunicator__registry)
        self.assertNotIn("Test.exchange_object.3", Test._MiddlewareCommunicator__registry)

    def test_get_communicators(self):
        """
        Test the get communicators functionality of the middleware. When the ``get_communicators`` method is called, the
        method should return a list of all available middleware wrappers.
        """
        import wrapyfi.tests.tools.class_test as class_test
        # importlib.reload(class_test)
        Test = class_test.Test

        if self.MWARE not in Test.get_communicators():
            self.skipTest(f"{self.MWARE} not installed")

        self.assertIn(self.MWARE, Test.get_communicators())

    def tearDown(self):
        from wrapyfi.connect.wrapper import MiddlewareCommunicator
        MiddlewareCommunicator.close_all_instances()


class ROS2TestWrapper(ZeroMQTestWrapper):
    """
    Test the ROS 2 wrapper. This test class inherits from the ZeroMQ test class, so all tests from the ZeroMQ test class
    are also run for the ROS 2 wrapper.
    """
    MWARE = "ros2"


class YarpTestWrapper(ZeroMQTestWrapper):
    """
    Test the YARP wrapper. This test class inherits from the ZeroMQ test class, so all tests from the ZeroMQ test class
    are also run for the YARP wrapper.
    """
    MWARE = "yarp"


class ROSTestWrapper(ZeroMQTestWrapper):
    """
    Test the ROS wrapper. This test class inherits from the ZeroMQ test class, so all tests from the ZeroMQ test class
    are also run for the ROS wrapper.
    """
    MWARE = "ros"


if __name__ == '__main__':
    unittest.main()



import logging
import os
from glob import glob

from wrapyfi.utils import SingletonOptimized, dynamic_module_import


class PublisherWatchDog(metaclass=SingletonOptimized):
    """
    A watchdog that scans for publishers and removes them from the ring if they are not established.
    """
    def __init__(self, repeats: int = 10, inner_repeats: int = 10):
        """
        Initialize the PublisherWatchDog.

        :param repeats: int: The number of times to repeat the scan
        param inner_repeats: int: The number of times to repeat the scan for each publisher
        """
        self.repeats = repeats
        self.inner_repeats = inner_repeats
        self.publisher_ring = []

    def add_publisher(self, publisher):
        """
        Add a publisher to the ring.

        :param publisher: Publisher: The publisher to add
        """
        self.publisher_ring.append(publisher)

    def remove_publisher(self, publisher):
        """
        Remove a publisher from the ring.

        :param publisher: Publisher: The publisher to remove
        """
        self.publisher_ring.remove(publisher)

    def scan(self):
        """
        Scan for publishers and remove them from the ring if they are established.
        """
        repeats = self.repeats
        while self.publisher_ring and (repeats > 0 | repeats <= -1):
            repeats -= 1
            for publisher in self.publisher_ring:
                found_publisher = publisher.establish(repeats=self.inner_repeats)
                if found_publisher:
                    self.publisher_ring.remove(publisher)


class Publishers(object):
    """
    A class that holds all publishers and their corresponding middleware communicators.
    """
    registry = {}
    mwares = set()

    @classmethod
    def register(cls, data_type: str, communicator: str):
        """
        Register a publisher for a given data type and middleware communicator.

        :param data_type: str: The data type to register the publisher for e.g., "NativeObject", "Image", "AudioChunk", etc.
        :param communicator: str: The middleware communicator to register the publisher for e.g., "ros", "ros2", "yarp", "zeromq", etc.
        :return: Callable[..., Any]: A decorator function that registers the decorated class as a publisher for the given data type and middleware communicator
        """
        def decorator(cls_):
            cls.registry[data_type + ":" + communicator] = cls_
            cls.mwares.add(communicator)
            return cls_
        return decorator

    @staticmethod
    def scan():
        """
        Scan for publishers and add them to the registry.
        """
        modules = glob(os.path.join(os.path.dirname(__file__), "..", "publishers", "*.py"), recursive=True)
        modules = ["wrapyfi.publishers." + module.replace(os.path.dirname(__file__) + "/../publishers/", "") for module in
                   modules]
        dynamic_module_import(modules, globals())


class Publisher(object):
    """
    A base class for all publishers.
    """
    def __init__(self, name: str, out_topic: str, carrier: str = "", should_wait: bool = True, **kwargs):
        """
        Initialize the Publisher.

        :param name: str: The name of the publisher
        :param out_topic: str: The name of the output topic
        :param carrier: str: The name of the carrier to use
        :param should_wait: bool: Whether to wait for the publisher to be established or not
        """
        self.__name__ = name
        self.out_topic = out_topic
        self.carrier = carrier
        self.should_wait = should_wait
        self.established = False

    def check_establishment(self, established: bool):
        """
        Check if the publisher is established and remove it from the ring if it is.

        :param established: bool: Whether the publisher is established or not
        :return: bool: Whether the publisher is established or not
        """
        if established:
            self.established = True
            if not self.should_wait:
                PublisherWatchDog().remove_publisher(self)
        elif not self.should_wait:
            PublisherWatchDog().scan()
            if self in PublisherWatchDog().publisher_ring:
                established = False
            else:
                established = True
        return established

    def establish(self, repeats: int = -1, **kwargs):
        """
        Establish the publisher.
        """
        raise NotImplementedError

    def publish(self, obj):
        """
        Publish an object.
        """
        raise NotImplementedError

    def close(self):
        """
        Close the connection.
        """
        raise NotImplementedError



"""
A message publisher and listener for native Python objects and Pint Quantities.

This script demonstrates the capability to transmit native Python objects and Pint Quantities using
the MiddlewareCommunicator within the Wrapyfi library. The communication follows the PUB/SUB pattern
allowing message publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using the NativeObject message
    - Transmitting a nested dummy Python object with native objects and Pint Quantities
    - Applying the PUB/SUB pattern with mirroring

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - Pint: Used for handling physical quantities with units

    Install using pip:
        ``pip install pint``

Run:
    # On machine 1 (or process 1): Publisher waits for keyboard input and transmits message
        ``python3 pint_example.py --mode publish``

    # On machine 2 (or process 2): Listener waits for message and prints the entire dummy object
        ``python3 pint_example.py --mode listen``
"""

import argparse

try:
    import pint
except ImportError:
    print("Install Pint before running this script.")

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class Notifier(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "NativeObject", "$mware", "Notifier", "/notify/test_pint_exchange",
        carrier="tcp", should_wait=True
    )
    def exchange_object(self, mware=None):
        """Exchange messages with Pint Quantities and other native Python objects."""
        msg = input("Type your message: ")

        # Creating a Pint Quantity
        ureg = pint.UnitRegistry()
        quantity = 42 * ureg.parse_expression('meter')

        # Constructing the message object to be transmitted
        ret = [{"message": msg,
                "pint_quantity": quantity,
                "list": [1, 2, 3]},
               "string",
               0.4344,
               {"other": (1, 2, 3, 4.32,)}]
        return ret,


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="A message publisher and listener for native Python objects and Pint Quantities.")
    parser.add_argument(
        "--mode", type=str, default="publish",
        choices={"publish", "listen"},
        help="The transmission mode"
    )
    parser.add_argument(
        "--mware", type=str, default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission"
    )
    return parser.parse_args()


def main(args):
    """Main function to initiate Notifier class and communication."""
    notifier = Notifier()
    notifier.activate_communication(Notifier.exchange_object, mode=args.mode)

    while True:
        msg_object, = notifier.exchange_object(mware=args.mware)
        print("Method result:", msg_object)


if __name__ == "__main__":
    args = parse_args()
    main(args)


import logging
import json
from typing import Optional, Literal, Tuple

import numpy as np
import yarp

from wrapyfi.connect.servers import Server, Servers
from wrapyfi.middlewares.yarp import YarpMiddleware
from wrapyfi.encoders import JsonEncoder, JsonDecodeHook


class YarpServer(Server):

    def __init__(self, name: str, out_topic: str, carrier: Literal["tcp", "udp", "mcast"] = "tcp",
                 out_topic_connect: Optional[str] = None, persistent: bool = True,
                 yarp_kwargs: Optional[dict] = None, **kwargs):
        """
        Initialize the server.

        :param name: str: Name of the server
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param out_topic_connect: str: Name of the output topic connection alias '/' (e.g. '/topic:out') to connect to.
                                        None appends ':out' to the out_topic. Default is None
        :param yarp_kwargs: dict: Additional kwargs for  the Yarp middleware
        :param kwargs: dict: Additional kwargs for the server
        """
        super().__init__(name, out_topic, carrier=carrier, out_topic_connect=out_topic_connect, **kwargs)
        YarpMiddleware.activate(**yarp_kwargs or {})
        self.style = yarp.ContactStyle()
        self.style.persistent = persistent
        self.style.carrier = self.carrier

        self.persistent = persistent

    def close(self):
        """
        Close the server.
        """
        if hasattr(self, "_port") and self._port:
            if self._port is not None:
                self._port.close()

    def __del__(self):
        self.close()


@Servers.register("NativeObject", "yarp")
class YarpNativeObjectServer(YarpServer):

    def __init__(self, name: str, out_topic: str, carrier: Literal["tcp", "udp", "mcast"] = "tcp",
                 out_topic_connect: Optional[str] = None, persistent: bool = True,
                 serializer_kwargs: Optional[dict] = None, deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        Specific server handling native Python objects, serializing them to JSON strings for transmission.

        :param name: str: Name of the server
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param out_topic_connect: str: Name of the output topic connection alias '/' (e.g. '/topic:out') to connect to.
                                        None appends ':out' to the out_topic. Default is None
        :param persistent: bool: Whether the server port should remain connected after closure. Default is True
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, out_topic, carrier=carrier, out_topic_connect=out_topic_connect, persistent=persistent, **kwargs)
        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        self._port = self._netconnect = None

    def establish(self):
        """
        Establish the connection to the server.
        """
        self._port = yarp.RpcServer()
        self._port.open(self.out_topic)
        if self.style.persistent:
            self._netconnect = yarp.Network.connect(self.out_topic, self.out_topic_connect, self.style)
        else:
            self._netconnect = yarp.Network.connect(self.out_topic, self.out_topic_connect, self.carrier)

        self._netconnect = yarp.Network.connect(self.out_topic, self.out_topic_connect, self.carrier)
        if self.persistent:
            self.established = True

    def await_request(self, *args, **kwargs):
        """
        Await and deserialize the client's request, returning the extracted arguments and keyword arguments.
        The method blocks until a message is received, then attempts to deserialize it using the configured JSON decoder
        hook, returning the extracted arguments and keyword arguments.

        :return: Tuple[list, dict]: A tuple containing two items:
                 - A list of arguments extracted from the received message
                 - A dictionary of keyword arguments extracted from the received message
        """
        if not self.established:
            self.establish()
        try:
            obj_msg = yarp.Bottle()
            obj_msg.clear()
            request = False
            while not request:
                request = self._port.read(obj_msg, True)
            [args, kwargs] = json.loads(obj_msg.get(0).asString(), object_hook=self._plugin_decoder_hook, **self._deserializer_kwargs)
            return args, kwargs
        except Exception as e:
            logging.error("[YARP] Service call failed: %s" % e)
            return [], {}

    def reply(self, obj):
        """
        Serialize the provided Python object to a JSON string and send it as a reply to the client.
        The method uses the configured JSON encoder for serialization before sending the resultant string to the client.

        :param obj: Any: The Python object to be serialized and sent
        """
        obj_str = json.dumps(obj, cls=self._plugin_encoder, **self._plugin_kwargs,
                             serializer_kwrags=self._serializer_kwargs)
        obj_msg = yarp.Bottle()
        obj_msg.clear()
        obj_msg.addString(obj_str)
        if self.persistent:
            self._port.reply(obj_msg)
        else:
            self._port.replyAndDrop(obj_msg)


@Servers.register("Image", "yarp")
class YarpImageServer(YarpNativeObjectServer):
    def __init__(self, name: str, out_topic: str, carrier: Literal["tcp", "udp", "mcast"] = "tcp",
                 out_topic_connect: Optional[str] = None, persistent: bool = True,
                 width: int = -1, height: int = -1, rgb: bool = True, fp: bool = False,
                 deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        Specific server handling image data as numpy arrays, serializing them to JSON strings for transmission.

        :param name: str: Name of the server
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param out_topic_connect: str: Name of the output topic connection alias '/' (e.g. '/topic:out') to connect to.
                                        None appends ':out' to the out_topic. Default is None
        :param persistent: bool: Whether the server port should remain connected after closure. Default is True
        :param width: int: Width of the image. Default is -1 (use the width of the received image)
        :param height: int: Height of the image. Default is -1 (use the height of the received image)
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        if "jpg" in kwargs:
            logging.warning("[YARP] YARP currently does not support JPG encoding in REQ/REP. Using raw image.")
            kwargs.pop("jpg")
        super().__init__(name, out_topic, carrier=carrier, out_topic_connect=out_topic_connect, persistent=persistent, deserializer_kwargs=deserializer_kwargs, **kwargs)
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp

    def reply(self, img: np.ndarray):
        """
        Serialize the provided image data and send it as a reply to the client.

        :param img: np.ndarray: Image to send formatted as a cv2 image - np.ndarray[img_height, img_width, channels]
        """
        if 0 < self.width != img.shape[1] or 0 < self.height != img.shape[0] or \
                not ((img.ndim == 2 and not self.rgb) or (img.ndim == 3 and self.rgb and img.shape[2] == 3)):
            raise ValueError("Incorrect image shape for publisher")
        # img = np.require(img, dtype=self._type, requirements='C')
        super().reply(img)


@Servers.register("AudioChunk", "yarp")
class YarpAudioChunkServer(YarpNativeObjectServer):
    def __init__(self, name: str, out_topic: str, carrier: Literal["tcp", "udp", "mcast"] = "tcp",
                 out_topic_connect: Optional[str] = None, persistent: bool = True,
                 channels: int = 1, rate: int = 44100, chunk: int = -1,
                 deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        Specific server handling audio data as numpy arrays, serializing them to JSON strings for transmission.

        :param name: str: Name of the server
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param out_topic_connect: str: Name of the output topic connection alias '/' (e.g. '/topic:out') to connect to.
                                        None appends ':out' to the out_topic. Default is None
        :param persistent: bool: Whether the server port should remain connected after closure. Default is True
        :param channels: int: Number of channels in the audio. Default is 1
        :param rate: int: Sampling rate of the audio. Default is 44100
        :param chunk: int: Number of samples in the audio chunk. Default is -1 (use the chunk size of the received audio)
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, out_topic, carrier=carrier, out_topic_connect=out_topic_connect, persistent=persistent, deserializer_kwargs=deserializer_kwargs, **kwargs)
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

    def reply(self, aud: Tuple[np.ndarray, int]):
        """
        Serialize the provided audio data and send it as a reply to the client.

        :param aud: Tuple[np.ndarray, int]: Audio chunk to publish formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        aud, rate = aud
        if aud is None:
            return
        if 0 < self.rate != rate:
            raise ValueError("Incorrect audio rate for server reply")
        chunk, channels = aud.shape if len(aud.shape) > 1 else (aud.shape[0], 1)
        self.chunk = chunk if self.chunk == -1 else self.chunk
        self.channels = channels if self.channels == -1 else self.channels
        if 0 < self.chunk != chunk or 0 < self.channels != channels:
            raise ValueError("Incorrect audio shape for publisher")
        aud = np.require(aud, dtype=np.float32, requirements='C')
        super().reply((chunk, channels, rate, aud))



from collections import Counter
from functools import reduce
import argparse
import json

import zmq

from wrapyfi.middlewares.zeromq import ZeroMQMiddlewareParamServer

DEFAULT_COMMAND = "set /foo/bar/42"


def access_nested_dict(d, keys):
    return reduce(lambda x,y: x[y], keys[:-1], d)[keys[-1]]


def parse_prefix(param_full: str, topics: dict):
    # split the string by '/' to get the individual components
    components = param_full.split('/')
    # If there is only one component in the list, add it to the topics dictionary
    # as a value and return
    if len(components) == 1:
        topics[""] = components[0]
        return topics
    # Get the first component and remove it from the list
    key = components.pop(0)
    # If the key doesn't exist in the topics dictionary, add it as a key
    # and recursively call parse_prefix with the remaining components
    if key not in topics:
        topics[key] = parse_prefix("/".join(components), {})
    # If the key already exists in the topics dictionary, recursively call
    # parse_prefix with the remaining components and update the value of the
    # key in the topics dictionary
    else:
        topics[key] = parse_prefix("/".join(components), topics[key])
    # return the updated topics dictionary
    return topics


def reverse_parse_prefix(topics: dict, prefix: str = ""):
    # create an empty list to store the generated strings
    strings = []
    # iterate over the keys in the topics dictionary
    for key in topics:
        # If the value of the key is a string, add the key and value to the list
        # of strings as a '/' separated string
        if isinstance(topics[key], str):
            strings.append(key + '/' + topics[key])
        # If the value of the key is a dictionary, recursively call generate_strings
        # with the value and add the returned list of strings to the list of strings
        # with the key as a prefix
        else:
            nested_strings = reverse_parse_prefix(topics[key])
            for string in nested_strings:
                strings.append(key + '/' + string)
    # return the list of strings filtered for repeated /'s
    return [prefix + string.replace('//', '/') for string in strings]


def main(role, param_ip, param_pub_port, param_sub_port, param_reqrep_port, param_poll_interval, param_poll_repeat, **kwargs):
    if role == "server":
        param_pub_address = f"tcp://{param_ip}:{param_pub_port}"
        param_sub_address = f"tcp://{param_ip}:{param_sub_port}"
        param_reqrep_address = f"tcp://{param_ip}:{param_reqrep_port}"

        ZeroMQMiddlewareParamServer.activate(**{"param_pub_address": param_pub_address,
                                                "param_sub_address": param_sub_address,
                                                "param_reqrep_address": param_reqrep_address,
                                                "param_poll_interval": param_poll_interval,
                                                "proxy_broker_spawn": "process", "verbose": True}, **kwargs)
        while True:
            pass

    elif role == "client":
        param_pub_address = f"tcp://{param_ip}:{param_pub_port}"
        param_reqrep_address = f"tcp://{param_ip}:{param_reqrep_port}"
        default_command = DEFAULT_COMMAND

        context = zmq.Context.instance()

        request_server = context.socket(zmq.REQ)
        request_server.connect(param_reqrep_address)

        param_server = context.socket(zmq.SUB)
        param_server.connect(param_pub_address)
        param_server.setsockopt(zmq.RCVTIMEO, 2000)

        while True:
            # send a request to the request server starting with write, read, delete, set or get
            new_commands = str(input(f"input command: (default - {default_command})")) or default_command
            default_command = new_commands
            # pass a list of commands to the request server e.g. ['set /foo/bar/21', 'set /foo/bar/baz/42', 'get /foo/bar', 'delete /foo/bar/baz', 'read /foo']
            if '[' in new_commands:
                new_commands = new_commands.replace('[\'', '').replace('\']', '')
                new_commands = new_commands.split('\', \'')
            # write supports writing full tree dicts e.g. write {'foo': {'bar': {'': '42', 'car': '43'}}}
            elif 'write' in new_commands:
                new_commands = new_commands.replace('write ', '')
                if new_commands.startswith('{'):
                    new_commands = new_commands.replace('\n', '').replace('\t', '')
                    new_commands = json.loads(new_commands)
                    new_commands = reverse_parse_prefix(new_commands, prefix="set ")
            # pass commands directly to the request server e.g. set /foo/bar/42
            else:
                new_commands = [new_commands]

            for new_command in new_commands:
                print("Sending request ", new_command)
                request_server.send_string(new_command)
                reply = request_server.recv_string()
                print("Received reply from server: %s" % reply)
                if "success:::" in reply:
                    topics = {}
                    current_prefix = reply.split(":::")[1].encode('utf-8')
                    param_server.subscribe(current_prefix)
                    # time.sleep(0.2)
                    prev_params = Counter()
                    param = "!"
                    while True:
                        # receive updates from the parameter server
                        try:
                            prefix, param, value = param_server.recv_multipart()
                        except zmq.error.Again:
                            print("No new parameters received. Need atleast one topic to subscribe to.")
                            break
                        # construct the full parameter name with the namespace prefix
                        prefix, param, value = prefix.decode('utf-8'), param.decode('utf-8'), value.decode('utf-8')
                        if (param in prev_params or param is None) and prev_params[param] == param_poll_repeat:
                            break
                        prev_params[param] += 1
                        full_param = "/".join([prefix, param, value])
                        parse_prefix(full_param, topics)
                        # print("Received update: %s" % (full_param))
                    try:
                        topic_results = access_nested_dict(topics, current_prefix.decode('utf-8').split('/'))
                    except KeyError:
                        print(current_prefix.decode('utf-8') + " has no children")
                        continue
                    print(json.dumps({current_prefix.decode('utf-8'): topic_results}, indent=None, default=str))
                    print("Reverse parse (always in set mode regardless of the transmitted command):")
                    print(reverse_parse_prefix(topic_results, prefix=f"set {current_prefix.decode('utf-8')}/"))
                    # close the connection
                    param_server.unsubscribe(current_prefix)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--param_ip", type=str, default="127.0.0.1", help="Parameter sever IP address")
    parser.add_argument("--param_pub_port", type=int, default=5655, help="Socket publishing (PUB) port")
    parser.add_argument("--param_sub_port", type=int, default=5656, help="Socket subscription (SUB) port")
    parser.add_argument("--param_reqrep_port", type=int, default=5659, help="Socket request (REQ)/reply (REP) port")
    parser.add_argument("--param_poll_interval", type=int, default=1, help="PUB/SUB poll interval in milliseconds "
                                                                           "(only used when role=server)")
    parser.add_argument("--param_poll_repeat", type=int, default=5, help="SUB poll repetition in terms of no. of "
                                                                         "attempts (only used when role=client)")
    parser.add_argument("--role", type=str, default="server", choices=["server", "client"],
                        help="Parameter server in serving (server) or requesting (client) mode")
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    main(**vars(args))



import time


from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class Test(MiddlewareCommunicator):

    @MiddlewareCommunicator.register("NativeObject", "$mware", "Test", "$topic",
                                     should_wait="$should_wait")
    def exchange_object(self, msg=None, mware=DEFAULT_COMMUNICATOR, topic="/test/test_native_exchange", should_wait=False):
        ret = {"message": msg,
               "set": {'a', 1, None},
               "list": [[[3, [4], 5.677890, 1.2]]],
               "string": "string of characters",
               "2": 2.73211,
               "dict": {"other": [None, False, 16, 4.32,]}}
        return ret,


def test_func(queue_buffer, mode="listen", mware=DEFAULT_COMMUNICATOR, topic="/test/test_native_exchange", iterations=2,
              should_wait=False):
    test = Test()
    test.activate_communication(test.exchange_object, mode=mode)
    for i in range(iterations):
        my_message = test.exchange_object(msg=f"signal_idx:{i}", mware=mware, topic=topic, should_wait=should_wait)
        if my_message is not None:
            print(f"result {mode}:", my_message[0]["message"])
            queue_buffer.put(my_message[0])
        time.sleep(0.5)



import logging
import os
from glob import glob
from typing import Optional

from wrapyfi.utils import dynamic_module_import


class Servers(object):
    """
    A class that holds all servers and their corresponding middleware communicators.
    """
    registry = {}
    mwares = set()

    @classmethod
    def register(cls, data_type: str, communicator: str):
        """
        Register a server with the given data type and middleware communicator.

        :param data_type: str: The data type to register the server for e.g., "NativeObject", "Image", "AudioChunk", etc.
        :param communicator: str: The middleware communicator to register the server for e.g., "ros", "ros2", "yarp", "zeromq", etc.
        :return: Callable: A decorator that registers the server with the given data type and middleware communicator
        """
        def decorator(cls_):
            cls.registry[data_type + ":" + communicator] = cls_
            cls.mwares.add(communicator)
            return cls_
        return decorator

    @staticmethod
    def scan():
        """
        Scan for servers and add them to the registry.
        """
        modules = glob(os.path.join(os.path.dirname(__file__), "..", "servers", "*.py"), recursive=True)
        modules = ["wrapyfi.servers." + module.replace(os.path.dirname(__file__) + "/../servers/", "") for module in
                   modules]
        dynamic_module_import(modules, globals())


class Server(object):
    """
    A base class for servers.
    """
    def __init__(self, name: str, out_topic: str, carrier: str = "", out_topic_connect: Optional[str] = None, **kwargs):
        """
        Initialize the server.

        :param name: str: The name of the server
        :param out_topic: str: The topic to publish to
        :param carrier: str: The middleware carrier to use
        :param out_topic_connect: str: The topic to connect to (this is deprecated and will be removed in the future since its usage is limited to YARP)
        """
        self.__name__ = name
        self.out_topic = out_topic
        self.carrier = carrier
        self.out_topic_connect = out_topic + ":out" if out_topic_connect is None else out_topic_connect
        self.established = False

    def establish(self):
        """
        Establish the server.
        """
        raise NotImplementedError

    def await_request(self, *args, **kwargs):
        """
        Await a request from a client.
        """
        raise NotImplementedError

    def reply(self, obj):
        """
        Reply to a client request.
        """
        raise NotImplementedError

    def close(self):
        """
        Close the connection.
        """
        raise NotImplementedError



"""
Encoder and Decoder for PaddlePaddle Tensor Data via Wrapyfi.

This script provides mechanisms to encode and decode PaddlePaddle tensor data using Wrapyfi.
It utilizes base64 encoding to convert binary data into ASCII strings.

The script contains a class, `PaddleTensor`, registered as a plugin to manage the
conversion of PaddlePaddle tensor data (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - PaddlePaddle: An open-source deep learning platform developed by Baidu (refer to https://www.paddlepaddle.org.cn/en/install/quick for installation instructions)
        Note: If PaddlePaddle is not available, HAVE_PADDLE will be set to False and
        the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install paddlepaddle``  # Basic installation of PaddlePaddle
"""

import io
import base64
from functools import lru_cache

import numpy as np

from wrapyfi.utils import *


try:
    import paddle
    from paddle import Tensor
    HAVE_PADDLE = True
except ImportError:
    HAVE_PADDLE = False


def paddle_device_to_str(device):
    """
    Convert a PaddlePaddle device to a string representation.

    :param device: Union[str, paddle.fluid.libpaddle.Place, dict]: Various possible types representing a PaddlePaddle device
    :return: Union[str, dict]: A string or dictionary representing the PaddlePaddle device
    """
    if device is None:
        return paddle.get_device()
    elif isinstance(device, dict):
        device_rets = {}
        for k, v in device.items():
            device_rets[paddle_device_to_str(k)] = paddle_device_to_str(v)
        return device_rets
    elif isinstance(device, paddle.fluid.libpaddle.Place):
        return str(device)[6:-1].replace("cuda", "gpu")
    elif isinstance(device, str):
        return device.replace("cuda", "gpu")
    else:
        raise ValueError(f'Unknown device type {type(device)}')


@lru_cache(maxsize=None)
def paddle_str_to_device(device):
    """
    Convert a string to a PaddlePaddle device.

    :param device: str: A string representing a PaddlePaddle device
    :return: paddle.fluid.libpaddle.Place: A PaddlePaddle place representing the device
    """
    if device is None:
        return paddle.get_device()
    elif isinstance(device, paddle.fluid.libpaddle.Place):
        return device
    elif isinstance(device, str):
        try:
            device_type, device_id = device.split(':')
        except ValueError:
            device_type = device
        return paddle.device._convert_to_place(device_type.replace("cuda", "gpu"))
    else:
        raise ValueError(f'Unknown device type {type(device)}')


@PluginRegistrar.register(types=None if not HAVE_PADDLE else paddle.Tensor.__mro__[:-1])
class PaddleTensor(Plugin):
    def __init__(self, load_paddle_device=None, map_paddle_devices=None, **kwargs):
        """
        Initialize the PaddleTensor plugin.

        :param load_paddle_device: Union[paddle.fluid.libpaddle.Place, str]: Default PaddlePaddle device to load tensors onto
        :param map_paddle_devices: dict: A dictionary mapping encoded device strings to decoding devices
        """
        self.map_paddle_devices = map_paddle_devices or {}
        if load_paddle_device is not None:
            self.map_paddle_devices['default'] = load_paddle_device
        self.map_paddle_devices = paddle_device_to_str(self.map_paddle_devices)

    def encode(self, obj, *args, **kwargs):
        """
        Encode PaddlePaddle tensor data into a base64 ASCII string.

        :param obj: paddle.Tensor: The PaddlePaddle tensor data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name, encoded data string, and device string
        """
        with io.BytesIO() as memfile:
            paddle.save(obj, memfile)
            obj_data = base64.b64encode(memfile.getvalue()).decode('ascii')
        obj_device = paddle_device_to_str(obj.place)
        return True, dict(__wrapyfi__=(str(self.__class__.__name__), obj_data, obj_device))

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a base64 ASCII string back into PaddlePaddle tensor data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the encoded data string and device string
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, paddle.Tensor]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - paddle.Tensor: The decoded PaddlePaddle tensor data
        """
        with io.BytesIO(base64.b64decode(obj_full[1].encode('ascii'))) as memfile:
            obj_device = self.map_paddle_devices.get(obj_full[2], self.map_paddle_devices.get('default', None))
            if obj_device is not None:
                obj_device = paddle_str_to_device(obj_device)
                return True, paddle.Tensor(paddle.load(memfile), place=obj_device)
            else:
                return True, paddle.load(memfile)




"""
Encoder and Decoder for Astropy Table Data via Wrapyfi.

This script provides a mechanism to encode and decode Astropy table data using
Wrapyfi. It makes use of base64 encoding to transform binary data into ASCII strings.

The script contains a class, `AstropyData`, registered as a plugin to handle the
conversion of Astropy data (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - Astropy: A library for Astronomy computations and data handling in Python.
        Note: If Astropy is not available, HAVE_ASTROPY will be set to False and
        the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install astropy``
"""

import io
import base64

from wrapyfi.utils import *

try:
    from astropy.table import Table
    HAVE_ASTROPY = True
except ImportError:
    HAVE_ASTROPY = False


@PluginRegistrar.register(types=None if not HAVE_ASTROPY else Table.__mro__[:-1])
class AstropyData(Plugin):
    def __init__(self, **kwargs):
        """
        Initialize the AstropyData plugin for encoding and decoding Astropy table data.

        This class provides methods to convert Astropy table data to an encoded form
        and decode it back to its original form. The data is encoded using base64
        to transform binary data into ASCII strings.
        """
        pass

    def encode(self, obj, *args, **kwargs):
        """
        Encode Astropy table data into a base64 ASCII string.

        :param obj: Table: The Astropy table data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name and the encoded data string
        """
        memfile = io.BytesIO()
        obj.write(memfile, format='fits')
        memfile.seek(0)
        obj_data = base64.b64encode(memfile.getvalue()).decode('ascii')
        memfile.close()
        return True, dict(__wrapyfi__=(str(self.__class__.__name__), obj_data))

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a base64 ASCII string back into Astropy table data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the encoded data string
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, astropy.Table]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - Table: The decoded Astropy table data
        """
        encoded_str = obj_full[1]
        if isinstance(encoded_str, str):
            encoded_str = encoded_str.encode('ascii')
        with io.BytesIO(base64.b64decode(encoded_str)) as memfile:
            memfile.seek(0)
            obj = Table.read(memfile, format='fits')
        return True, obj



import logging
import sys
import json
import queue
import time
import os
from typing import Optional, Any

import numpy as np
import cv2
import rospy
import rostopic
import std_msgs.msg
import sensor_msgs.msg

from wrapyfi.connect.listeners import Listener, Listeners, ListenerWatchDog
from wrapyfi.middlewares.ros import ROSMiddleware
from wrapyfi.encoders import JsonDecodeHook


QUEUE_SIZE = int(os.environ.get("WRAPYFI_ROS_QUEUE_SIZE", 5))
WATCHDOG_POLL_REPEAT = None


class ROSListener(Listener):

    def __init__(self, name: str, in_topic: str, carrier: str = "tcp", should_wait: bool = True,
                 queue_size: int = QUEUE_SIZE, ros_kwargs: Optional[dict] = None, **kwargs):
        """
        Initialize the subscriber.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param queue_size: int: Size of the queue for the subscriber. Default is 5
        :param ros_kwargs: dict: Additional kwargs for the ROS middleware
        :param kwargs: dict: Additional kwargs for the subscriber
        """
        if carrier or carrier != "tcp":
            logging.warning("[ROS] ROS does not support other carriers than TCP for PUB/SUB pattern. Using TCP.")
            carrier = "tcp"
        super().__init__(name, in_topic, carrier=carrier, should_wait=should_wait, **kwargs)
        ROSMiddleware.activate(**ros_kwargs or {})
        
        self.queue_size = queue_size

    def close(self):
        """
        Close the subscriber
        """
        if hasattr(self, "_subscriber") and self._subscriber:
            if self._subscriber is not None:
                self._subscriber.unregister()

    def __del__(self):
        self.close()


@Listeners.register("NativeObject", "ros")
class ROSNativeObjectListener(ROSListener):

    def __init__(self, name: str, in_topic: str, carrier: str = "tcp", should_wait: bool = True, queue_size: int =QUEUE_SIZE,
                 deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        The NativeObject listener using the ROS String message assuming the data is serialized as a JSON string.
        Deserializes the data (including plugins) using the decoder and parses it to a Python object.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param queue_size: int: Size of the queue for the subscriber. Default is 5
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, in_topic, carrier=carrier, should_wait=should_wait, queue_size=queue_size, **kwargs)

        self._subscriber = self._queue = None

        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        ListenerWatchDog().add_listener(self)

    def establish(self):
        """
        Establish the subscriber.
        """
        self._queue = queue.Queue(maxsize=0 if self.queue_size is None or self.queue_size <= 0 else self.queue_size)
        self._subscriber = rospy.Subscriber(self.in_topic, std_msgs.msg.String, callback=self._message_callback)
        self.established = True

    def listen(self) -> Any:
        """
        Listen for a message.

        :return: Any: The received message as a native python object
        """
        if not self.established:
            self.establish()
        try:
            obj_str = self._queue.get(block=self.should_wait)
            return json.loads(obj_str, object_hook=self._plugin_decoder_hook, **self._deserializer_kwargs)
        except queue.Empty:
            return None

    def _message_callback(self, msg):
        """
        Callback for the subscriber.

        :param msg: std_msgs.msg.String: The received message
        """
        try:
            self._queue.put(msg.data, block=False)
        except queue.Full:
            logging.warning(f"[ROS] Discarding data because listener queue is full: {self.in_topic}")


@Listeners.register("Image", "ros")
class ROSImageListener(ROSListener):

    def __init__(self, name: str, in_topic: str, carrier: str = "tcp", should_wait: bool = True, queue_size: int = QUEUE_SIZE,
                 width: int = -1, height: int = -1, rgb: bool = True, fp: bool = False, jpg: bool = False, **kwargs):
        """
        The Image listener using the ROS Image message parsed to a numpy array.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param queue_size: int: Size of the queue for the subscriber. Default is 5
        :param width: int: Width of the image. Default is -1 (use the width of the received image)
        :param height: int: Height of the image. Default is -1 (use the height of the received image)
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be decompressed from JPG. Default is False
        """
        super().__init__(name, in_topic, carrier=carrier, should_wait=should_wait, queue_size=queue_size, **kwargs)
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        if self.fp:
            self._encoding = '32FC3' if self.rgb else '32FC1'
            self._type = np.float32
        else:
            self._encoding = 'bgr8' if self.rgb else 'mono8'
            self._type = np.uint8
        if self.jpg:
            self._encoding = 'jpeg'
            self._type = np.uint8

        self._pixel_bytes = (3 if self.rgb else 1) * np.dtype(self._type).itemsize

        self._subscriber = self._queue = None

        ListenerWatchDog().add_listener(self)

    def establish(self):
        """
        Establish the subscriber.
        """
        self._queue = queue.Queue(maxsize=0 if self.queue_size is None or self.queue_size <= 0 else self.queue_size)
        if self.jpg:
            self._subscriber = rospy.Subscriber(self.in_topic, sensor_msgs.msg.CompressedImage, callback=self._message_callback)
        else:
            self._subscriber = rospy.Subscriber(self.in_topic, sensor_msgs.msg.Image, callback=self._message_callback)
        self.established = True

    def listen(self):
        """
        Listen for a message.

        :return: np.ndarray: The received message as a numpy array formatted as a cv2 image np.ndarray[img_height, img_width, channels]
        """
        if not self.established:
            self.establish()
        try:
            if self.jpg:
                format, data = self._queue.get(block=self.should_wait)
                if format != 'jpeg':
                    raise ValueError(f"Unsupported image format: {format}")
                if self.rgb:
                    img = cv2.imdecode(np.frombuffer(data, np.uint8), cv2.IMREAD_COLOR)
                else:
                    img = cv2.imdecode(np.frombuffer(data, np.uint8), cv2.IMREAD_GRAYSCALE)
            else:
                height, width, encoding, is_bigendian, data = self._queue.get(block=self.should_wait)
                if encoding != self._encoding:
                    raise ValueError("Incorrect encoding for listener")
                if 0 < self.width != width or 0 < self.height != height or len(data) != height * width * self._pixel_bytes:
                    raise ValueError("Incorrect image shape for listener")
                img = np.frombuffer(data, dtype=np.dtype(self._type).newbyteorder('>' if is_bigendian else '<')).reshape((height, width, -1))
                if img.shape[2] == 1:
                    img = img.squeeze(axis=2)
            return img
        except queue.Empty:
            return None

    def _message_callback(self, data):
        """
        Callback for the subscriber.

        :param data: sensor_msgs.msg.Image: The received message
        """
        try:
            if self.jpg:
                self._queue.put((data.format, data.data), block=False)
            else:
                self._queue.put((data.height, data.width, data.encoding, data.is_bigendian, data.data), block=False)
        except queue.Full:
            logging.warning(f"[ROS] Discarding data because listener queue is full: {self.in_topic}")


@Listeners.register("AudioChunk", "ros")
class ROSAudioChunkListener(ROSListener):

    def __init__(self, name: str, in_topic: str, carrier: str = "tcp", should_wait: bool = True, queue_size: int = QUEUE_SIZE,
                 channels: int = 1, rate: int = 44100, chunk: int = -1, **kwargs):
        """
        The AudioChunk listener using the ROS Image message parsed to a numpy array.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param queue_size: int: Size of the queue for the subscriber. Default is 5
        :param channels: int: Number of channels in the audio. Default is 1
        :param rate: int: Sampling rate of the audio. Default is 44100
        :param chunk: int: Number of samples in the audio chunk. Default is -1 (use the chunk size of the received audio)
        """
        super().__init__(name, in_topic, carrier=carrier, should_wait=should_wait, queue_size=queue_size,
                         width=chunk, height=channels, rgb=False, fp=True, jpg=False, **kwargs)
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

        self._subscriber = self._queue = None
        ListenerWatchDog().add_listener(self)

    def establish(self):
        """
        Establish the subscriber.
        """
        try:
            from wrapyfi_ros_interfaces.msg import ROSAudioMessage
        except ImportError:
            import wrapyfi
            logging.error("[ROS] Could not import ROSAudioMessage. "
                          "Make sure the ROS messages in wrapyfi_extensions/wrapyfi_ros_interfaces are compiled. "
                          "Refer to the documentation for more information: \n" +
                          wrapyfi.__doc__ + "ros_interfaces_lnk.html")
            sys.exit(1)
        self._queue = queue.Queue(maxsize=0 if self.queue_size is None or self.queue_size <= 0 else self.queue_size)
        self._subscriber = rospy.Subscriber(self.in_topic, ROSAudioMessage, callback=self._message_callback)
        self.established = True

    def listen(self):
        """
        Listen for a message.

        :return: Tuple[np.ndarray, int]: The received message as a numpy array formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        if not self.established:
            self.establish()
        try:
            chunk, channels, rate, encoding, is_bigendian, data = self._queue.get(block=self.should_wait)
            if 0 < self.rate != rate:
                raise ValueError("Incorrect audio rate for listener")
            if encoding not in ['S16LE', 'S16BE']:
                raise ValueError("Incorrect encoding for listener")
            if 0 < self.chunk != chunk or self.channels != channels or len(data) != chunk * channels * 4:
                raise ValueError("Incorrect audio shape for listener")
            aud = np.frombuffer(data, dtype=np.dtype(np.float32).newbyteorder('>' if is_bigendian else '<')).reshape((chunk, channels))
            # aud = aud / 32767.0
            return aud, rate
        except queue.Empty:
            return None, self.rate

    def _message_callback(self, data):
        """
        Callback for the subscriber.

        :param data: wrapyfi_ros_interfaces.msg.ROSAudioMessage: The received message
        """
        try:
            self._queue.put((data.chunk_size, data.channels, data.sample_rate, data.encoding, data.is_bigendian, data.data), block=False)
        except queue.Full:
            logging.warning(f"[ROS] Discarding data because listener queue is full: {self.in_topic}")


@Listeners.register("Properties", "ros")
class ROSPropertiesListener(ROSListener):
    """
    Gets rospy parameters. Behaves differently from other data types by directly acquiring ROS parameters.
    Note that the listener is not guaranteed to receive the updated signal, since the listener can trigger before
    property is set. The property decorated method returns accept native python objects (excluding None),
    but care should be taken when using dictionaries, since they are analogous with node namespaces:
    http://wiki.ros.org/rospy/Overview/Parameter%20Server
    """
    def __init__(self, name: str, in_topic: str, carrier: str = "tcp", should_wait: bool = True, queue_size: int = QUEUE_SIZE, **kwargs):
        """
        The PropertiesListener using the ROS Parameter Server.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether the subscriber should wait for a parameter to be set. Default is True
        :param queue_size: int: Size of the queue for the subscriber. Default is 5
        """
        super().__init__(name, in_topic, carrier=carrier, should_wait=should_wait, queue_size=queue_size, **kwargs)
        self._subscriber = self._queue = None

        if not self.should_wait:
            ListenerWatchDog().add_listener(self)

        self.previous_property = False

    def await_connection(self, in_topic: Optional[int] = None, repeats: Optional[int] = None):
        """
        Wait for a parameter to be set.

        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param repeats: int: Number of times to check for the parameter. None for infinite. Default is None
        """
        connected = False
        if in_topic is None:
            in_topic = self.in_topic
        logging.info(f"[ROS] Waiting for property: {in_topic}")
        if repeats is None:
            if self.should_wait:
                repeats = -1
            else:
                repeats = 1

            while repeats > 0 or repeats <= -1:
                repeats -= 1
                self.previous_property = rospy.get_param(self.in_topic, False)
                connected = True if self.previous_property else False
                if connected:
                    logging.info(f"[ROS] Found property: {in_topic}")
                    break
                time.sleep(0.2)
        return connected

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the subscriber.

        :param repeats: int: Number of times to check for the parameter. None for infinite. Default is None
        """
        established = self.await_connection(repeats=repeats)
        return self.check_establishment(established)

    def listen(self):
        """
        Listen for a message.

        :return: Any: The received message as a native python object
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                obj = None
            else:
                obj = self.previous_property
            return obj
        else:
            obj = rospy.get_param(self.in_topic, False)
            return obj


@Listeners.register("ROSMessage", "ros")
class ROSMessageListener(ROSListener):

    def __init__(self, name: str, in_topic: str, carrier: str = "tcp", should_wait: bool = True, queue_size: int = QUEUE_SIZE, **kwargs):
        """
        The ROSMessageListener using the ROS message type inferred from the message type. Supports standard ROS msgs.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether the subscriber should wait for a message to be published. Default is True
        :param queue_size: int: Size of the queue for the subscriber. Default is 5
        """
        super().__init__(name, in_topic, carrier=carrier, should_wait=should_wait, queue_size=queue_size, **kwargs)
        self._subscriber = self._queue = self._topic_type = None

        ListenerWatchDog().add_listener(self)

    def establish(self):
        """
        Establish the subscriber.
        """
        self._queue = queue.Queue(maxsize=0 if self.queue_size is None or self.queue_size <= 0 else self.queue_size)
        self._topic_type, topic_str, _ = rostopic.get_topic_class(self.in_topic, blocking=self.should_wait)
        if self._topic_type is None:
            return
        self._subscriber = rospy.Subscriber(self.in_topic, self._topic_type, callback=self._message_callback)
        self.established = True

    def listen(self):
        """
        Listen for a message.

        :return: rospy.msg: The received message as a ROS message object
        """
        if not self.established:
            self.establish()
        try:
            obj = self._queue.get(block=self.should_wait)

            return obj  # self._topic_type.deserialize_numpy(obj_str)
        except queue.Empty:
            return None

    def _message_callback(self, msg):
        """
        Callback for the subscriber.

        :param msg: rospy.msg: The received message as a ROS message object
        """
        try:
            self._queue.put(msg, block=False)
        except queue.Full:
            logging.warning(f"[ROS] Discarding data because listener queue is full: {self.in_topic}")




import logging
import sys
import json
import time
import threading
import os
import importlib.util
import queue
from typing import Optional, Tuple

import numpy as np
import cv2
import rclpy
from rclpy.node import Node
import std_msgs.msg
import sensor_msgs.msg

from wrapyfi.connect.servers import Server, Servers
from wrapyfi.middlewares.ros2 import ROS2Middleware
from wrapyfi.encoders import JsonEncoder, JsonDecodeHook


class ROS2Server(Server, Node):

    def __init__(self, name: str, out_topic: str, ros2_kwargs: Optional[dict] = None, **kwargs):
        """
        Initialize the server.

        :param name: str: Name of the server
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param ros2_kwargs: dict: Additional kwargs for the ROS 2 middleware
        :param kwargs: dict: Additional kwargs for the server
        """
        carrier = "tcp"
        if "carrier" in kwargs and kwargs["carrier"] not in ["", None]:
            logging.warning(
                "[ROS 2] ROS 2 currently does not support explicit carrier setting for REQ/REP pattern. Using TCP.")
        if "carrier" in kwargs:
            del kwargs["carrier"]

        ROS2Middleware.activate(**ros2_kwargs or {})
        Server.__init__(self, name, out_topic, carrier=carrier, **kwargs)
        Node.__init__(self, name + str(hex(id(self))))

    def close(self):
        """
        Close the server.
        """
        if hasattr(self, "_server") and self._server:
            if self._server is not None:
                self.destroy_node()
        if hasattr(self, "_background_callback") and self._background_callback:
            if self._background_callback is not None:
                self._background_callback.join()

    def __del__(self):
        self.close()


@Servers.register("NativeObject", "ros2")
class ROS2NativeObjectServer(ROS2Server):
    SEND_QUEUE = queue.Queue(maxsize=1)
    RECEIVE_QUEUE = queue.Queue(maxsize=1)

    def __init__(self, name: str, out_topic: str,
                 serializer_kwargs: Optional[dict] = None, deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        Specific server handling native Python objects, serializing them to JSON strings for transmission.

        :param name: str: Name of the server
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, out_topic, **kwargs)
        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        self._server = None

    def establish(self):
        """
        Establish the connection to the server
        """
        try:
            from wrapyfi_ros2_interfaces.srv import ROS2NativeObjectService
        except ImportError:
            import wrapyfi
            logging.error("[ROS 2] Could not import ROS2NativeObjectService. "
                          "Make sure the ROS 2 services in wrapyfi_extensions/wrapyfi_ros2_interfaces are compiled. "
                          "Refer to the documentation for more information: \n" +
                          wrapyfi.__doc__ + "ros2_interfaces_lnk.html")
            sys.exit(1)

        self._server = self.create_service(ROS2NativeObjectService, self.out_topic, self._service_callback)

        self._req_msg = ROS2NativeObjectService.Request()
        self._rep_msg = ROS2NativeObjectService.Response()
        self.established = True

    def await_request(self, *args, **kwargs):
        """
        Await and deserialize the client's request, returning the extracted arguments and keyword arguments.
        The method blocks until a message is received, then attempts to deserialize it using the configured JSON decoder
        hook, returning the extracted arguments and keyword arguments.

        :return: Tuple[list, dict]: A tuple containing two items:
                 - A list of arguments extracted from the received message
                 - A dictionary of keyword arguments extracted from the received message
        """
        if not self.established:
            self.establish()
        try:
            self._background_callback = threading.Thread(name='ros2_server', target=rclpy.spin_once,
                                                         args=(self,), kwargs={})
            self._background_callback.setDaemon(True)
            self._background_callback.start()

            request = ROS2NativeObjectServer.RECEIVE_QUEUE.get(block=True)
            [args, kwargs] = json.loads(request.request, object_hook=self._plugin_decoder_hook, **self._deserializer_kwargs)
            return args, kwargs
        except Exception as e:
            logging.error("[ROS 2] Service call failed %s" % e)
            return [], {}

    @staticmethod
    def _service_callback(request, _response):
        """
        Callback for the ROS 2 service.

        :param request: ROS2NativeObjectService.Request: The request message
        :param _response: ROS2NativeObjectService.Response: The response message
        """
        ROS2NativeObjectServer.RECEIVE_QUEUE.put(request)
        return ROS2NativeObjectServer.SEND_QUEUE.get(block=True)

    def reply(self, obj):
        """
        Serialize the provided Python object to a JSON string and send it as a reply to the client.
        The method uses the configured JSON encoder for serialization before sending the resultant string to the client.

        :param obj: Any: The Python object to be serialized and sent
        """
        try:
            obj_str = json.dumps(obj, cls=self._plugin_encoder, **self._plugin_kwargs,
                                 serializer_kwrags=self._serializer_kwargs)
            self._rep_msg.response = obj_str
            ROS2NativeObjectServer.SEND_QUEUE.put(self._rep_msg, block=False)
        except queue.Full:
            logging.warning(f"[ROS 2] Discarding data because queue is full. "
                            f"This happened due to bad synchronization in {self.__name__}")


@Servers.register("Image", "ros2")
class ROS2ImageServer(ROS2Server):
    SEND_QUEUE = queue.Queue(maxsize=1)
    RECEIVE_QUEUE = queue.Queue(maxsize=1)

    def __init__(self, name: str, out_topic: str,
                 width: int = -1, height: int = -1, rgb: bool = True, fp: bool = False, jpg: bool = False,
                 deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        Specific server handling native Python objects, serializing them to JSON strings for transmission.

        :param name: str: Name of the server
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param width: int: Width of the image. Default is -1 (use the width of the received image)
        :param height: int: Height of the image. Default is -1 (use the height of the received image)
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be decompressed from JPG. Default is False
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, out_topic, **kwargs)
        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        if self.fp:
            self._encoding = '32FC3' if self.rgb else '32FC1'
            self._type = np.float32
        else:
            self._encoding = 'bgr8' if self.rgb else 'mono8'
            self._type = np.uint8
        if self.jpg:
            self._encoding = 'jpeg'
            self._type = np.uint8

        self._server = None

    def establish(self):
        """
        Establish the connection to the server.
        """
        try:
            from wrapyfi_ros2_interfaces.srv import ROS2ImageService, ROS2CompressedImageService
        except ImportError:
            import wrapyfi
            logging.error("[ROS 2] Could not import ROS2NativeObjectService. "
                          "Make sure the ROS 2 services in wrapyfi_extensions/wrapyfi_ros2_interfaces are compiled. "
                          "Refer to the documentation for more information: \n" +
                          wrapyfi.__doc__ + "ros2_interfaces_lnk.html")
            sys.exit(1)
        if self.jpg:
            self._server = self.create_service(ROS2CompressedImageService, self.out_topic, self._service_callback)
            self._req_msg = ROS2CompressedImageService.Request()
            self._rep_msg = ROS2CompressedImageService.Response()
        else:
            self._server = self.create_service(ROS2ImageService, self.out_topic, self._service_callback)
            self._req_msg = ROS2ImageService.Request()
            self._rep_msg = ROS2ImageService.Response()
        self.established = True

    def await_request(self, *args, **kwargs):
        """
        Await and deserialize the client's request, returning the extracted arguments and keyword arguments.
        The method blocks until a message is received, then attempts to deserialize it using the configured JSON decoder
        hook, returning the extracted arguments and keyword arguments.

        :return: Tuple[list, dict]: A tuple containing two items:
                 - A list of arguments extracted from the received message
                 - A dictionary of keyword arguments extracted from the received message
        """
        if not self.established:
            self.establish()
        try:
            self._background_callback = threading.Thread(name='ros2_server', target=rclpy.spin_once,
                                                         args=(self,), kwargs={})
            self._background_callback.setDaemon(True)
            self._background_callback.start()

            request = ROS2ImageServer.RECEIVE_QUEUE.get(block=True)
            [args, kwargs] = json.loads(request.request, object_hook=self._plugin_decoder_hook, **self._deserializer_kwargs)
            return args, kwargs
        except Exception as e:
            logging.error("[ROS 2] Service call failed %s" % e)
            return [], {}

    @staticmethod
    def _service_callback(request, _response):
        """
        Callback for the ROS 2 service.

        :param request: ROS2NativeObjectService.Request: The request message
        :param _response: ROS2NativeObjectService.Response: The response message
        :return: ROS2NativeObjectService.Response: The response message
        """
        ROS2ImageServer.RECEIVE_QUEUE.put(request)
        return ROS2ImageServer.SEND_QUEUE.get(block=True)

    def reply(self, img: np.ndarray):
        """
        Serialize the provided image data and send it as a reply to the client.

        :param img: np.ndarray: Image to send formatted as a cv2 image - np.ndarray[img_height, img_width, channels]
        """
        try:
            if 0 < self.width != img.shape[1] or 0 < self.height != img.shape[0] or \
                    not ((img.ndim == 2 and not self.rgb) or (img.ndim == 3 and self.rgb and img.shape[2] == 3)):
                raise ValueError("Incorrect image shape for publisher")
            img = np.require(img, dtype=self._type, requirements='C')
            img_msg = self._rep_msg.response
            if self.jpg:
                img_msg.header.stamp = rclpy.clock.Clock().now().to_msg()
                img_msg.format = "jpeg"
                img_msg.data = np.array(cv2.imencode('.jpg', img)[1]).tobytes()
            else:
                img_msg.header.stamp = rclpy.clock.Clock().now().to_msg()
                img_msg.height = img.shape[0]
                img_msg.width = img.shape[1]
                img_msg.encoding = self._encoding
                img_msg.is_bigendian = img.dtype.byteorder == '>' or (img.dtype.byteorder == '=' and sys.byteorder == 'big')
                img_msg.step = img.strides[0]
                img_msg.data = img.tobytes()
            self._rep_msg.response = img_msg
            ROS2ImageServer.SEND_QUEUE.put(self._rep_msg, block=False)
        except queue.Full:
            logging.warning(f"[ROS 2] Discarding data because queue is full. "
                            f"This happened due to bad synchronization in {self.__name__}")


@Servers.register("AudioChunk", "ros2")
class ROS2AudioChunkServer(ROS2Server):
    SEND_QUEUE = queue.Queue(maxsize=1)
    RECEIVE_QUEUE = queue.Queue(maxsize=1)

    def __init__(self, name: str, out_topic: str,
                 channels: int = 1, rate: int = 44100, chunk: int = -1,
                 deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        Specific server handling audio data as numpy arrays.

        :param name: str: Name of the server
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param channels: int: Number of channels in the audio. Default is 1
        :param rate: int: Sampling rate of the audio. Default is 44100
        :param chunk: int: Number of samples in the audio chunk. Default is -1 (use the chunk size of the received audio)
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, out_topic, **kwargs)
        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        self.channels = channels
        self.rate = rate
        self.chunk = chunk

    def establish(self):
        """
        Establish the connection to the server.
        """
        try:
            from wrapyfi_ros2_interfaces.srv import ROS2AudioService
        except ImportError:
            import wrapyfi
            logging.error("[ROS 2] Could not import ROS2AudioService. "
                          "Make sure the ROS 2 services in wrapyfi_extensions/wrapyfi_ros2_interfaces are compiled. "
                          "Refer to the documentation for more information: \n" +
                          wrapyfi.__doc__ + "ros2_interfaces_lnk.html")
            sys.exit(1)
        self._server = self.create_service(ROS2AudioService, self.out_topic, self._service_callback)
        self._req_msg = ROS2AudioService.Request()
        self._rep_msg = ROS2AudioService.Response()
        self.established = True

    def await_request(self, *args, **kwargs):
        """
        Await and deserialize the client's request, returning the extracted arguments and keyword arguments.
        The method blocks until a message is received, then attempts to deserialize it using the configured JSON decoder
        hook, returning the extracted arguments and keyword arguments.

        :return: Tuple[list, dict]: A tuple containing two items:
                 - A list of arguments extracted from the received message
                 - A dictionary of keyword arguments extracted from the received message
        """
        if not self.established:
            self.establish()
        try:
            self._background_callback = threading.Thread(name='ros2_server', target=rclpy.spin_once,
                                                         args=(self,), kwargs={})
            self._background_callback.setDaemon(True)
            self._background_callback.start()

            request = ROS2AudioChunkServer.RECEIVE_QUEUE.get(block=True)
            [args, kwargs] = json.loads(request.request, object_hook=self._plugin_decoder_hook, **self._deserializer_kwargs)
            return args, kwargs
        except Exception as e:
            logging.error("[ROS 2] Service call failed %s" % e)
            return [], {}

    @staticmethod
    def _service_callback(request, _response):
        """
        Callback for the ROS 2 service.

        :param request: ROS2NativeObjectService.Request: The request message
        :param _response: ROS2NativeObjectService.Response: The response message
        :return: ROS2NativeObjectService.Response: The response message
        """
        ROS2AudioChunkServer.RECEIVE_QUEUE.put(request)
        return ROS2AudioChunkServer.SEND_QUEUE.get(block=True)

    def reply(self, aud: Tuple[np.ndarray, int]):
        """
        Serialize the provided audio data and send it as a reply to the client.

        :param aud: Tuple[np.ndarray, int]: Audio chunk to publish formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        try:
            aud, rate = aud
            if aud is None:
                return
            if 0 < self.rate != rate:
                raise ValueError("Incorrect audio rate for publisher")
            chunk, channels = aud.shape if len(aud.shape) > 1 else (aud.shape[0], 1)
            self.chunk = chunk if self.chunk == -1 else self.chunk
            self.channels = channels if self.channels == -1 else self.channels
            if 0 < self.chunk != chunk or 0 < self.channels != channels:
                raise ValueError("Incorrect audio shape for publisher")
            aud = np.require(aud, dtype=np.float32, requirements='C')

            aud_msg = self._rep_msg.response
            aud_msg.header.stamp = self.get_clock().now().to_msg()
            aud_msg.chunk_size = chunk
            aud_msg.channels = channels
            aud_msg.sample_rate = rate
            aud_msg.is_bigendian = aud.dtype.byteorder == '>' or (aud.dtype.byteorder == '=' and sys.byteorder == 'big')
            aud_msg.encoding = 'S16BE' if aud_msg.is_bigendian else 'S16LE'
            aud_msg.step = aud.strides[0]
            aud_msg.data = aud.tobytes()  # (aud * 32767.0).tobytes()
            self._rep_msg.response = aud_msg
            ROS2AudioChunkServer.SEND_QUEUE.put(self._rep_msg, block=False)
        except queue.Full:
            logging.warning(f"[ROS 2] Discarding data because queue is full. "
                            f"This happened due to bad synchronization in {self.__name__}")
